/*
 * Task1Test.java
 * JUnit based test
 *
 */
package cz.muni.fi.pb138.jaro2019.task1;

import java.io.IOException;
import java.net.URI;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import junit.framework.TestCase;
import org.w3c.dom.Document;

/**
 *
 * @author originally Petr Adamek 2008, 
 * changes by Tomas Pitner, Feb 2011, Feb 2013, Feb 2014, Jan 2015, Jan 2017, 
 * Feb 2018, Mar 2019
 */
public class Task1Test extends TestCase {
    
    private static final String COMPANYXML = "company.xml";
    public static final String FIRSTNOTE = "//person[@pid='1']/note[1]";
    public static final String SECONDNOTE = "//person[@pid='1']/note[2]";
    private static final String SALARY_SUM_EMPLOYEE = "sum(//employee/salary)";
    private static final String EMPLOYEE_COUNT = "count(//employee/salary)";

    private Main solution;
    private XPath xpath;
    private static final double DELTA = 0.001;

    public Task1Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
        URI uri = Task1Test.class.getResource(COMPANYXML).toURI();
        if (uri == null) {
            throw new IOException("The XML file for testing cannot be found.");
        }
        solution = Main.newInstance(uri);
        xpath = XPathFactory.newInstance().newXPath();
    }

    public void testAvgEmployeeSalary() throws XPathExpressionException {
        final Document document = solution.getDocument();
        double actualAvg = solution.getAvgEmployeeSalary();
        assertEquals(38693.1875, actualAvg, DELTA);
    }

    public void testIncreaseEmployeesSalary() throws XPathExpressionException {
        final Document document = solution.getDocument();

        solution.increaseEmployeesSalary(30000.0, 20.0);

        double avgAfter = (Double)xpath.evaluate(
                SALARY_SUM_EMPLOYEE, document, XPathConstants.NUMBER) 
                / (Double) xpath.evaluate(
                EMPLOYEE_COUNT, document, XPathConstants.NUMBER);
        
        assertEquals(41189.325, avgAfter, DELTA);
    }

    public void testAddNote() throws XPathExpressionException {
        final Document document = solution.getDocument();
        
        double origNotesCount = (Double)xpath.evaluate(
                "count(//note)", document, XPathConstants.NUMBER);
        
        String note = (String)xpath.evaluate(
                FIRSTNOTE, document, XPathConstants.STRING);
        assertEquals("Poznámka", note);
        
        solution.addNote("1", "Nová poznámka");
        
        note = (String)xpath.evaluate(
                FIRSTNOTE, document, XPathConstants.STRING);
        assertEquals("Poznámka", note);
        
        String note2 = (String)xpath.evaluate(
                SECONDNOTE, document, XPathConstants.STRING);
        assertEquals("Nová poznámka", note2);
        
        assertEquals(origNotesCount+1, (Double)xpath.evaluate(
                "count(//note)", document, XPathConstants.NUMBER), DELTA);
    }
}