package cz.muni.fi.pb138.jaro2019.task1;

import javax.xml.xpath.XPathFactory;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * ***********************************************************************
 * Take this skeleton code and implement the missing bodies of
 * Task 1.A <code>increaseEmployeesSalary<code>getAvgEmployeeSalary<code>addNote methods.
 *
 * After completion, the <code>main method should produce output similar to:
 * Running demo...

 * Task 1.B Average salary: 38693.1875

 * Task 1.A Salary of persons lower or equal to 30000.0 increased by 20.0%

 * Task 1.B Average salary now: 41189.325

 * Task 1.C Person with pid=1 is now:
 * <?xml version="1.0" encoding="UTF-8"?><person pid="1"> 	
 *   <!-- name is required, non-empty string -->	 
 *   <name>Pavel</name>	 
 *   <!-- surname is required, non-empty string -->
 *   <surname>Novák</surname> 
 *   <!-- note optional element, repeatable, contains any string -->  
 *   <note>Poznámka</note>   
 *   <note>New note to person 1</note></person>
 * Demo finished, see the above output to check whether it is correct.
 *
 */
public class Main {

    /**
     * W3C object model representation of a XML document. Note: We use the
     * interface(!) not its implementation
     */
    private final Document doc;

    /**
     * Task 1, part A: Complete this method. You are likely to consult the Java
     * Core API documentation for org.w3c.dom package when working on this task.
     * You can also use the XPath sub-API (javax.xml.xpath) to evaluate XPath
     * expressions from within Java methods thought this is not needed.
     *
     * Hint for those not experienced in Java: you can use double
     * Double.parseDouble (String s) to convert the text node inside the salary
     * element to a double value. You can write any other (private) method you
     * need.
     *
     * You can suppose that the input document is always valid, i.e. its
     * structure fulfills the criteria specified in the comments inside the
     * document company.xml.
     *
     * @param upperBound double number indicating the current salary - the new
     * salary will only be increased for employees (not heads) 
     * with the current salary lower or equal to upperBound.
     * @param percent percentual increase, eg. 10 means the new salary will be
     * 110 % of the current one.
     *
     * @return the List of "employee" elements where the salary has been
     * increased by the method. May be empty of there were no such employees.
     */
    public List<Element> increaseEmployeesSalary(double upperBound, double percent) {
        List<Element> increasedSalaryEmployees = new ArrayList<>();
        NodeList employees = doc.getElementsByTagName("employee");
        for (int i = 0; i < employees.getLength(); i++) {
            Element employee = (Element) (employees.item(i));
            Element salary = (Element) employee.getElementsByTagName("salary").item(0);
            increaseSalary(salary, upperBound, percent);
            increasedSalaryEmployees.add(employee);
        }
        return increasedSalaryEmployees;
    }

    private void increaseSalary(Element salaryElement, double upperBound, double percent) {
        double origSalary = Double.parseDouble(salaryElement.getTextContent());
        if(origSalary <= upperBound) {
            salaryElement.setTextContent(Double.toString((percent + 100) / 100 * origSalary));
        }
    }

    /**
     * Task 1, part B: Complete this method. You are likely to consult the Java
     * Core API documentation for org.w3c.dom package when working on this task.
     *
     * You can also use the XPath sub-API (javax.xml.xpath) to evaluate XPath
     * expressions from within Java methods.
     *
     * You can suppose that the input document is always valid, i.e. its
     * structure fulfills the criteria specified in the comments inside the
     * document company.xml.
     *
     * @return average salary of all employees (not heads) 
     * including heads in the company as double.
     */
    public double getAvgEmployeeSalary() {
        double avgSalaries = 0.0;

        try {
            XPathFactory xpf = XPathFactory.newInstance();
            XPath xPath = xpf.newXPath();
            avgSalaries
                    = (Double) xPath.evaluate(
                            "sum(//employee/salary) div count(//employee/salary)", 
                            doc, XPathConstants.NUMBER);
        } catch (XPathExpressionException ex) {
            Logger.getLogger(Main.class
                    .getName()).log(Level.SEVERE, "Error in XPath:", ex);
        }
        return avgSalaries;

    }

    /**
     * Task 1, part C: Complete this method. The method adds a new note to an
     * existing person identified by its person ID (pid) and returns this person
     * element. If such a person does not exist, the method does nothing and
     * returns null.
     *
     * @param pid of the person to add the note to this person. pid must not be
     * null nor empty (the method does not check it)
     * @param note the note to be added to the person with the given pid
     * @return the Element person with the given pid and new note added
     */
    public Element addNote(String pid, String note) {
        // complete the method
        NodeList people = doc.getElementsByTagName("person");
        for (int i = 0; i < people.getLength(); i++) {
            Element person = (Element) people.item(i);
            if (pid.equals(person.getAttribute("pid"))) {
                Element noteElement = doc.createElement("note");
                noteElement.setTextContent(note);
                person.appendChild(noteElement);
                return person;
            }
        }
        return null;
    }

    /**
     * **********************************************************************
     */
    /**
     * Create a new instance of this class and read the content of the XML file
     * under the given URL.
     */
    public static Main
            newInstance(URI uri) throws SAXException,
            ParserConfigurationException, IOException {
        return new Main(uri);

    }

    /**
     * Create a new instance of this class and read the content of the given XML
     * file.
     */
    public static Main
            newInstance(File file)
            throws SAXException, ParserConfigurationException, IOException {
        return newInstance(file.toURI());

    }

    /**
     * Constructor creating a new instance of Uloha1 class reading from the file
     * at the given URI
     */
    private Main(URI uri) throws SAXException, ParserConfigurationException,
            IOException {
        // Vytvorime instanci tovarni tridy
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        // Pomoci tovarni tridy ziskame instanci DocumentBuilderu
        DocumentBuilder builder = factory.newDocumentBuilder();
        // DocumentBuilder pouzijeme pro zpracovani XML dokumentu
        // a ziskame model dokumentu ve formatu W3C DOM
        doc
                = builder.parse(uri.toString());

    }

    public void serializetoXML(URI output)
            throws IOException, TransformerConfigurationException, TransformerException {
        // Vytvorime instanci tovarni tridy
        TransformerFactory factory = TransformerFactory.newInstance();
        // Pomoci tovarni tridy ziskame instanci tzv. kopirovaciho transformeru
        Transformer transformer = factory.newTransformer();
        // Vstupem transformace bude dokument v pameti
        DOMSource source = new DOMSource(doc);
        // Vystupem transformace bude vystupni soubor
        StreamResult result = new StreamResult(output.toString());
        // Provedeme transformaci
        transformer
                .transform(source, result);

    }

    public void serializetoXML(File output) throws IOException,
            TransformerException {
        serializetoXML(output.toURI());

    }

    public static void main(String[] args) throws IOException, SAXException, ParserConfigurationException, TransformerException, XPathExpressionException {

        System.out.println("Running demo...");

        if (args.length < 1) {
            System.err.println("Input file name is missing.");

            return;

        } else if (args.length < 2) {
            System.err.println("Output file name is missing.");

            return;

        }

        File input = new File(args[0]);
        File output = new File(args[1]);

        Main task1 = newInstance(input);

        TransformerFactory tf = TransformerFactory.newInstance();
        Transformer trans = tf.newTransformer();

        double upperBound = 30000.0;
        double percent = 20;
        System.out.println("\nTask 1.B Average salary: " + task1.getAvgEmployeeSalary());
        task1.increaseEmployeesSalary(upperBound, percent);
        System.out.println("\nTask 1.A Salary of persons lower or equal to "
                + upperBound + " increased by " + percent + "%");
        System.out.println("\nTask 1.B Average salary now: " + task1.getAvgEmployeeSalary());
        System.out.println("\nTask 1.C Person with pid=1 is now:");
        trans
                .transform(new DOMSource(task1.addNote("1", "New note to person 1")), new StreamResult(System.out));
        task1
                .serializetoXML(output);
        System.out.println("\nDemo finished, see the above output to check whether it is correct.");

    }

    /**
     * Help method for testing purposes.
     */
    Document getDocument() {
        return doc;
    }

}